#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cmath>
#include <ext/hash_map>
#include <vector>
#include "graph.h"
#include "auxiliar.h"

using namespace std;	

int main(int argc, char **argv){

  int k = atoi(argv[2]);  // Valor k
  int w = atoi(argv[3]);  // Valor w
  int c = atoi(argv[4]);  // valor c


  c=log2(c);              // Se calcula el logaritmo en base 2 
                          // para después dividir desplazando
 

  // #### Cálculo de la máxima clave y del máximo tamaño de tabla de dispersión ####

  int max_sum=w*w*(255>>c); // máximo valor para una componente de un pixel
  int bits_max_sum=0;
  int aux=max_sum;
  while(aux!=0){
    aux=aux>>1;
    ++bits_max_sum; // máximo número de bits
  } 
  int bits_max_sum_2=bits_max_sum*2;
 
  // tamaño máximo de tabla de dispersión
  int t_max_hash=(max_sum<<(bits_max_sum_2)) + (max_sum << bits_max_sum) + max_sum + 1;

  // #######################################################################
  

  
  std::vector<dist_ima> **tabla_hash=new std::vector<dist_ima> *[t_max_hash];
  for(int i=0;i<t_max_hash;++i){
    tabla_hash[i]=0;
  }
  

  // #### Calcular cuantas imágenes hay en el índice ####

  int num_imagenes=0;     // Valor número de imágenes 
  ifstream f_aux;
  char nombre_aux[30];
  f_aux.open(argv[1]);
  while(f_aux>>nombre_aux)
    ++num_imagenes;
  f_aux.close();

  // ##################################################


  ifstream f_ind, f_img;
  f_ind.open(argv[1]);
  char nombre[30];

  int *num_patches=new int[num_imagenes];
  int cont_ima=0;
 
  // Vector de distancias
  int *distancias=new int [num_imagenes*num_imagenes];
  for(int i=0;i<num_imagenes*num_imagenes;++i)
    distancias[i]=0;

  

  while(f_ind >> nombre){ 
    f_img.open(nombre);
    
    int width, height;
    int maxValue;
    char magic_number[30];
    
    f_img >> magic_number;
    f_img >> width >> height >> maxValue;
    
    int num_pixeles=width*height;
    pixel *pixeles=new pixel[num_pixeles];
    int length_inicio=f_img.tellg();
    f_img.close();
    f_img.open(nombre, ios::binary);


    // get length of file
    f_img.seekg(0,ios::end);
    int length_final=f_img.tellg();
    int length=length_final-length_inicio;
    f_img.seekg(length_inicio, ios::beg);


    // allocate memory
    char *buffer=new char[length];


    // read data as a block
    f_img.read(buffer, length);



    // ######## Dar formato al buffer #########

    int ind=0;
    for(int i=0;i<num_pixeles;++i){

      // Componente R
      
      char ch; 
      do{
	ch=buffer[ind];
	++ind;
      }while(ch==' ' || ch=='\n');

      int r=0;
      do{
	int const digitVal=ch-'0';
	r=r*10+digitVal;
	ch=buffer[ind];
	++ind;
      }while(ch>='0' && ch<='9');
      
      pixeles[i].r=r>>c;
      

      // Componente G
      
      do{
	ch=buffer[ind];
	++ind;
      }while(ch==' ' || ch=='\n');
      int g=0;
      do{
        int const digitVal=ch-'0';
	g=g*10+digitVal;
	ch=buffer[ind];
	++ind;
      }while(ch>='0' && ch<='9');
      
      pixeles[i].g=g>>c;
      

      // Componente B

      do{
	ch=buffer[ind];
	++ind;
      }while(ch==' ' || ch=='\n');
      int b=0;
      do{
	int const digitVal=ch-'0';
	b=b*10+digitVal;
	ch=buffer[ind];
	++ind;
      }while(ch>='0' && ch<='9');
      
      pixeles[i].b=b>>c;
    }
    delete [] buffer;

    // ########################################



    
    // #### Cálculo de elementos repetidos ####
    
    __gnu_cxx::hash_map<int, int, __gnu_cxx::hash<int>, eqstr> H; // tabla hash

    int sumR, sumB, sumG;
    int parches_alto=height-w+1;
    int parches_ancho=width-w+1;
    for(int i=0;i<parches_alto;++i){
      for(int j=0;j<parches_ancho;++j){
	sumR=sumG=sumB=0;
	for(int fil=0;fil<w;++fil){
	  int p=posicion(fil+i, j, width);
	  for(int col=0;col<w;++col){
	    sumR+=pixeles[col+p].r;
	    sumG+=pixeles[col+p].g;
	    sumB+=pixeles[col+p].b;
	  }
	}

	int clave=(sumR<<bits_max_sum_2) + (sumG<<bits_max_sum) + sumB; // clave

	if(H.find(clave)!= H.end()) 
	  H[clave]+=1;  // Si encuentra el elemento aumenta el número
	else
	  H[clave]=1; // Si no lo encuentra lo inserta con valor inicial 1
      }
    }

    delete [] pixeles;

    // ########################################


    
    num_patches[cont_ima]=(height-w+1)*(width-w+1); // número de parches

 

    // #### Cálculo de distancias ####

    __gnu_cxx::hash_map<int, int, __gnu_cxx::hash<int>, eqstr>::iterator it;
    

    /* Vector auxiliar para guardar las distancias
       Reduce el tiempo de acceso porque las distancias
       están en forma de vector y el acceso aleatorio
       tiene menos fallos en cache debido a que el vector
       es mucho más pequeño. */
    int *auxiliar=new int[num_imagenes];
    for(int j=0;j<num_imagenes;++j)
      auxiliar[j]=0;

    
    // Recorre todos los elementos de la tabla hash sin repetidos
    for(it=H.begin();it!=H.end();++it){
      int indice = it->first;

      // Si no hay ningún elemento insertado, crea un std vector 
      // para guardar los comunes
      if(!tabla_hash[indice]){ 
	tabla_hash[indice]=new std::vector<dist_ima>;
      }
      // Si ya está creado el vector calcula la distancia (como
      // el máximo valor que podrían dar si no fueran comunes 
      // menos lo que pierden por ser comunes) y posteriormente 
      // se resta a la distancia.
      else{
	std::vector<dist_ima>::iterator inicio=tabla_hash[indice]->begin();
	std::vector<dist_ima>::iterator final=tabla_hash[indice]->end();
	int cuenta1=it->second;
	while(inicio!=final){
	  int dist=cuenta1+inicio->cuenta-abs(cuenta1-inicio->cuenta);
	  auxiliar[inicio->ima]-=dist;
	  ++inicio;
	}
      }  
      dist_ima aux;
      aux.cuenta=it->second;
      aux.ima=cont_ima;
      tabla_hash[indice]->push_back(aux); // Siempre se inserta en la tabla hash
    }
    // Introduce los valores del vector auxiliar en el vector distancias
    for(int q=0;q<num_imagenes;++q){
      if(auxiliar[q]){
	int p=posicion(q, cont_ima, num_imagenes);
	distancias[p]+=auxiliar[q];
      }
    }

    // ##########################################################


    delete [] auxiliar; 
    ++cont_ima;
    f_img.close();
  }
  f_ind.close();


  // Replicar las distancias de la triangular superior
  // en la triangular inferior
  for(int i=0;i<num_imagenes;++i){
    for(int j=i+1;j<num_imagenes;++j){
      int pos1=posicion(i, j, num_imagenes);
      int pos2=posicion(j, i, num_imagenes);
      distancias[pos2]=distancias[pos1];
    }
  }


  // A cada posición se le suma el máximo valor posible de la distancia
  for(int i=0;i<num_imagenes;++i){
    for(int j=0;j<num_imagenes;++j){
      if(i!=j){
	int position=posicion(i, j, num_imagenes);
	int total_patches=num_patches[i]+num_patches[j];
	distancias[position]+=total_patches;
      }
    }
  }  


  // #### Insertar las k menores distancias en el grafo ####

  graph grafo_imagenes(num_imagenes);
  double aux_coste;
  int vertice_dest;
  for(int i=0;i<num_imagenes;++i){   // Bucle para insertar las aristas
    int *ya_elegido=new int[k];
    for(int p=0;p<k;++p)
      ya_elegido[p]=-1;
    for(int j=0;j<k;++j){
      aux_coste=INFINITY;
      vertice_dest=0;
      for(int q=0;q<i;++q){
	int position=posicion(i,q,num_imagenes);
	if(distancias[position]<aux_coste && no_esta(ya_elegido, k, q)){
	  aux_coste=distancias[position];
	  vertice_dest=q;
	}
      }
      for(int q=i+1;q<num_imagenes;++q){
	int position=posicion(i,q,num_imagenes);
	if(distancias[position]<aux_coste && no_esta(ya_elegido, k, q)){
	  aux_coste=distancias[position];
	  vertice_dest=q;
	}
      }
      if(aux_coste!=INFINITY){
	grafo_imagenes.insertar(i, vertice_dest, aux_coste);
	grafo_imagenes.insertar(vertice_dest, i, aux_coste);
	ya_elegido[j]=vertice_dest;
      }
    }
    delete [] ya_elegido;
  }
  // ##########################################################





  // #### Dijkstra ####
 
  for(int i=0;i<num_imagenes;++i) 
    grafo_imagenes.dijkstra(i); // distancia de una imagen a todas 
  

  grafo_imagenes.imprimir();    // imprimir distancias

  // #################


  delete [] distancias;

  return 0;
}
